import torch
import torch.nn as nn
from .positional import AbsolutePositionalEmbedding
from .mask import create_causal_mask


class ClassicTransformerBase(nn.Module):
    """Base class for Classic Transformer models - common logic for both decoders and encoders."""

    def __init__(
            self,
            embedding: nn.Embedding,
            layers: nn.ModuleList,
            absolute_embedding: AbsolutePositionalEmbedding = None,
            use_flash_attention: bool = False,
            use_relative_embedding: bool = False,
            *args,
            **kwargs,
    ):
        super(ClassicTransformerBase, self).__init__(*args, **kwargs)

        self.embedding = embedding
        self.pos_embedding = absolute_embedding
        self.use_flash_attention = use_flash_attention
        self.use_relative_embedding = use_relative_embedding

        self.layers = layers
        self.num_layers = len(layers) if layers else 0

    def update_max_len(self, max_seq_len: int):
        self.layers[0].update_max_len(max_seq_len)

    def moe_router_loss(self):
        return torch.stack([self.layers[i].moe_router_loss() for i in range(self.num_layers) if self.layers[i].use_moe or self.layers[i].use_moe_att]).mean()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Shared logic for encoders and decoders - apply embeddings and positional encoding
        x = self.embedding(x)
        if self.pos_embedding is not None:
            x = self.pos_embedding(x)
        return x


class ClassicTransformerDecoder(ClassicTransformerBase):
    """Classic Transformer decoder - for decoder-only Transformer models"""

    def __init__(self, embed_dim: int, vocab_size: int, use_head_norm: bool = False, init_identity_norm: bool = False, *args, **kwargs):
        super(ClassicTransformerDecoder, self).__init__(*args, **kwargs)

        self.head = nn.Linear(embed_dim, vocab_size)
        self.use_head_norm = use_head_norm
        if use_head_norm:
            self.head_norm = nn.LayerNorm(embed_dim)
            if init_identity_norm:
                self.head_norm.weight.data.fill_(1.0)
                self.head_norm.bias.data.fill_(0.0)
        else:
            self.head_norm = None

    def reset_self_attn_cache(self):
        for i in range(self.num_layers):
            self.layers[i].attention.reset_inner_cache()

    def forward(self, x: torch.Tensor, attention_mask: torch.Tensor = None, use_self_attn_cache: bool = False, current_positions: torch.Tensor = None) -> torch.Tensor:
        x = super().forward(x)  # apply embeddings
        seq_len = x.size(1)
        if not self.use_flash_attention and self.use_relative_embedding:
            mask = create_causal_mask(seq_len, device=x.device)
            if attention_mask is not None:
                mask &= attention_mask.unsqueeze(1).unsqueeze(1).bool()
        elif attention_mask is not None:
            mask = attention_mask.unsqueeze(1).unsqueeze(1).bool()
        else:
            mask = None

        # Process layers
        for i in range(self.num_layers):
            x = self.layers[i](x, mask=mask, use_self_attn_cache=use_self_attn_cache, current_positions=current_positions)
        return self.head(self.head_norm(x) if self.use_head_norm else x)


class ClassicTransformerEncoder(ClassicTransformerBase):
    """Classic Transformer encoder - for encoder-only Transformer models"""

    def forward(self, x: torch.Tensor, attention_mask: torch.Tensor = None) -> tuple[torch.Tensor, torch.Tensor]:
        x = super().forward(x)  # apply embeddings
        if attention_mask is not None:
            attention_mask = attention_mask.unsqueeze(1).unsqueeze(1).bool()

        hidden_states = []
        # Process own layers
        for i in range(self.num_own_layers):
            x = self.layers[i](x, mask=attention_mask)
            hidden_states.append(x)
        return x, torch.stack(hidden_states)



import torch
import torch.nn as nn
from typing import TypeAlias, Literal

ResidualGateType: TypeAlias = Literal['static', 'elementwise', 'linear']
SlotStatusType: TypeAlias = Literal['mean', 'linear']

class ResidualGate(nn.Module):
    def __init__(
            self,
            stm_size: int,
            embed_dim: int,
            use_gate: bool = False,
            gate_type: ResidualGateType = 'static',
            per_slot_gate: bool = True,
            init_gate: float = 0.0,
            use_tanh_gate: bool = True,
            slot_status_type: SlotStatusType = 'mean',
            disable_residual: bool = False,
            **kwargs,
    ):
        super(ResidualGate, self).__init__(**kwargs)
        self.use_gate = use_gate
        self.per_slot_gate = per_slot_gate
        self.gate_type = gate_type
        self.use_tanh_gate = use_tanh_gate
        self.slot_status_type = slot_status_type
        self.disable_residual = disable_residual

        if self.use_gate:
            if self.gate_type == 'linear':
                self.gate = nn.Linear(stm_size, stm_size if self.per_slot_gate else 1)
            else:
                gate_shape = (stm_size, 1) if self.per_slot_gate else (1,)
                self.gate = nn.Parameter(torch.full(gate_shape, init_gate))
        else:
            self.gate = None

        self.gate_activation = nn.Tanh() if self.use_tanh_gate else nn.Sigmoid()

        if self.slot_status_type == 'linear':
            assert self.gate_type == 'linear' or self.per_slot_gate == True, \
                'Cannot use linear slot status with not per-slot elementwise gate'
            self.slot_status = nn.Linear(embed_dim, 1)
        else:
            self.slot_status = None

    def _slot_statuses(self, updated_stm: torch.Tensor):
        if self.gate_type == 'linear':
            if self.slot_status_type == 'linear':
                return self.slot_status(updated_stm).squeeze(-1)
            else:
                return updated_stm.mean(dim=-1)
        else:
            if self.per_slot_gate and self.slot_status_type == 'linear':
                return self.slot_status(updated_stm)
            else:
                mean_dim = -1 if self.per_slot_gate else [1, 2]
                return updated_stm.mean(dim=mean_dim, keepdim=True)

    def _dynamic_gate(self, old_value: torch.Tensor, new_value: torch.Tensor):
        if self.gate_type == 'linear':
            statuses = self._slot_statuses(new_value + old_value)
            gate_input = self.gate(statuses).unsqueeze(-1)
        else:
            gate_input = self.gate * self._slot_statuses(new_value + old_value)
        return self.gate_activation(gate_input)

    def _calculate_output(self, layer_gate: torch.Tensor, old_value: torch.Tensor, new_value: torch.Tensor) -> torch.Tensor:
        if self.use_tanh_gate:
            return (1 + layer_gate) * new_value + (1 - layer_gate) * old_value
        else:
            return layer_gate * new_value + (1 - layer_gate) * old_value

    def forward(self, old_value: torch.Tensor, new_value: torch.Tensor) -> torch.Tensor:
        if self.disable_residual:
            return new_value
        if not self.use_gate:
            return new_value + old_value

        if self.gate_type == 'static':
            layer_gate = self.gate_activation(self.gate)
        else:
            layer_gate = self._dynamic_gate(old_value, new_value)

        return self._calculate_output(layer_gate, old_value, new_value)
